@@ -584,9 +584,9 @@ int ide_config_cb2 (const pci_config_t *config)
 {
 	ob_ide_init(config->path,
 		    config->assigned[0] & ~0x0000000F,
-		    config->assigned[1] & ~0x0000000F,
+		    (config->assigned[1] & ~0x0000000F) + 2,
 		    config->assigned[2] & ~0x0000000F,
-		    config->assigned[3] & ~0x0000000F);
+		    (config->assigned[3] & ~0x0000000F) + 2);
 	return 0;
 }
 
@@ -876,6 +876,9 @@ int i82378_config_cb(const pci_config_t *config)
 #ifdef CONFIG_DRIVER_PC_KBD
     ob_pc_kbd_init(config->path, "8042", arch->io_base, 0x60ULL, 0);
 #endif
+#ifdef CONFIG_DRIVER_IDE
+    ob_ide_init(config->path, 0x1f0, 0x3f6, 0x170, 0x376);
+#endif
 
     return 0;
 }
@@ -890,11 +893,13 @@ static void ob_pci_add_properties(phandle_t phandle,
 	int status,id;
 	uint16_t vendor_id, device_id;
 	uint8_t rev;
+	uint8_t class_prog;
 	uint32_t class_code;
 
 	vendor_id = pci_config_read16(addr, PCI_VENDOR_ID);
 	device_id = pci_config_read16(addr, PCI_DEVICE_ID);
 	rev = pci_config_read8(addr, PCI_REVISION_ID);
+	class_prog = pci_config_read8(addr, PCI_CLASS_PROG);
 	class_code = pci_config_read16(addr, PCI_CLASS_DEVICE);
 
     if (pci_dev) {
@@ -922,7 +927,7 @@ static void ob_pci_add_properties(phandle_t phandle,
 	set_int_property(dev, "vendor-id", vendor_id);
 	set_int_property(dev, "device-id", device_id);
 	set_int_property(dev, "revision-id", rev);
-	set_int_property(dev, "class-code", class_code << 8);
+	set_int_property(dev, "class-code", class_code << 8 | class_prog);
 
 	if (config->irq_pin) {
 		OLDWORLD(set_int_property(dev, "AAPL,interrupts",
